# -*- coding: utf-8 -*-
"""
Tools for interacting with Ed LMS.

..codeauthor Grzegorz Pawełczuk <grzegorz.pawelczuk@nftlearning.com>
"""
from typing import Optional

from nftl_ed_lms_tools.api.usergroups import Usergroups
from nftl_ed_lms_tools.api.users import Users
from nftl_ed_lms_tools.ed_api import OptStr


class EdClient:
    """
    Ed LMS API wrapper
    Partial implementation of methods that full documentation
    is available here http://developer.edapp.com/
    """

    def __init__(self, token, api_url=None):
        # type: (str, OptStr)-> None
        self._token = token
        self._api_url = api_url
        self._users_api: Optional[Users] = None
        self._usergroups_api: Optional[Usergroups] = None

    def get_users_api(self):
        # type: () -> Users
        """
        Provides Wrapper for methods available in Users category

        """
        if not self._users_api:
            self._users_api = Users(self._token, self._api_url)
        return self._users_api

    def get_usergroups_api(self):
        # type: () -> Usergroups
        """
        Provides Wrapper for methods available in User groups category

        """
        if not self._usergroups_api:
            self._usergroups_api = Usergroups(self._token, self._api_url)
        return self._usergroups_api



# -*- coding: utf-8 -*-
"""
Tools for interacting with Ed LMS API.

..codeauthor Grzegorz Pawełczuk <grzegorz.pawelczuk@nftlearning.com>
"""
import json
from enum import Enum
from json import JSONDecodeError

import requests
import logging as log

from abc import ABC
from typing import Any, Dict, Optional
from http.client import responses
from requests import Response, RequestException


class CallType(Enum):
    POST = 1
    GET = 2
    DELETE = 3


OptData = Optional[Dict[str, Any]]
OptStr = Optional[str]


class EdApi(ABC):
    """
    Ed API abstract handler

    Args:
        token: Ed admin bearer token to authorize API calls
        api_url: Ed API base Url
    """

    API_VERSION = "v1"

    def __init__(self, token, api_url):
        # type: (str, OptStr) -> None
        self.api_url = api_url
        if not api_url:
            self.api_url = 'https://rest.edapp.com/'
        self.token = token

    def _get_token(self, token):
        # type: (OptStr) -> str
        """
        Validate token - use provided if available or provide admins one

        Args:
            token: optional users auth token
        Returns:
            Auth token that scan be used
        """
        return token if token else self.token

    @staticmethod
    def _get_data(response):
        # type: (Response)-> OptData
        """
        Universal Response validator

        Args:
            response - :py:class:`requests.Response` object to be validated
        Returns:
            None if data is invalid or Dict[str, Any] when success
        """
        if response.ok:
            try:
                data = response.json()
                log.debug(f'ED response "{data}" from {response.url}')
                return data
            except JSONDecodeError as e:
                error_data = response.content or e
                if error_data:
                    log.error(
                        f'ED API error "{error_data}" from {response.url}')
        else:
            status = response.status_code
            name = responses[status]
            error: str = response.headers.get('x-amzn-errortype') or response.text
            msg = 'HTTP error ED (%s - %s, error: %s) from "%s"' % (
                status, name, error, response.url)
            log.error(msg)
        return None

    def _call(self, method, data=None, token=None, call_type=CallType.GET):
        # type: (str,Optional[Any],Optional[str],CallType) -> Optional[Any]
        """
        Wrapper for request builder

        Args:
            method: Slack API method name
            data: request parameters
            token: optional users auth token
        Returns:
            data structure that is available in json or None
            when error occurred
        """
        req = self._call_api(
            f'{self.API_VERSION}/%s' % method,
            data,
            token,
            call_type
        )

        if req is not None:
            return self._get_data(req)
        return None

    def _call_api(self, method, data, token=None, call_type=CallType.GET):
        # type: (str,OptData,OptStr,CallType) -> Optional[Response]
        """
        Slack POST request builder

        Args:
            method: Slack API method name
            data: request parameters
            token: optional users auth token
        Returns:
            :py:class:`requests.Response` for provided request
        """
        url = f'{self.api_url}{method}'
        try:
            if call_type == CallType.GET:
                log.info(f'ED Requesting {url} with GET')
                return requests.get(
                    url,
                    headers=self.get_headers(token)
                )
            elif call_type == CallType.POST:
                log.info(f'ED Requesting {url} with POST')
                return requests.post(
                    url,
                    headers=self.get_headers(token),
                    data=bytes(json.dumps(data), 'utf-8')
                )
            elif call_type == CallType.DELETE:
                log.info(f'ED Requesting {url} with DELETE')
                return requests.delete(
                    url,
                    headers=self.get_headers(token),
                    data=bytes(json.dumps(data), 'utf-8')
                )
        except RequestException as e:
            log.error(f'ED Error handling req. "{url}" with an exception: {e}')

        return None

    def get_headers(self, token=None):
        # type: (OptStr)-> Dict[str, str]
        return {
            'Authorization': 'Bearer %s' % self._get_token(token),
            'Content-type': 'application/json; charset=utf-8'
        }



# -*- coding: utf-8 -*-
"""
Tools for interacting with Ed API.

..codeauthor Grzegorz Pawełczuk <grzegorz.pawelczuk@nftlearning.com>
"""
import logging
from typing import Optional, Any, List, Dict

from nftl_ed_lms_tools.ed_api import EdApi, CallType


class Usergroups(EdApi):
    """
    Ed API wrapper that is handling user groups category
    """
    API_VERSION = "v1"

    def __init__(self, token: str, api_url: str = None) -> None:
        super().__init__(token, api_url)

    def get(self, token=None):
        # type: (Optional[str]) -> Optional[List[Any]]
        """
        Get a list of user groups.
        http://developer.edapp.com/#usergroups_get
        Args:
            token: optional auth token that will overwrite EdApi token
        Returns:
            None if data is invalid or user groups info as
            :py:class:`typing.List` when success
        """
        return super()._call(
            method='usergroups',
            token=token,
            call_type=CallType.GET
        )

    def create_or_update_group(self, group_data, token=None):
        # type: (Dict[str,Any], Optional[str]) -> Optional[Dict[str,Any]]
        """
        Create or update a user group.
        http://developer.edapp.com/#usergroups_post

        Args:
            group_data: group data to create or update
            token: optional auth token that will overwrite EdApi token
        Returns:
            None if data is invalid or group info as :py:class:`typing.Dict`
            when success
        """
        if 'id' not in group_data and 'id' not in group_data:
            msg = f'ED group "id" not provided in "{group_data}"'
            logging.error(msg)
            raise AttributeError(msg)

        return super()._call(
            method='usergroups',
            data=group_data,
            token=token,
            call_type=CallType.POST
        )




# -*- coding: utf-8 -*-
"""
Automated tests for Ed provisioner.

..codeauthor Grzegorz Pawełczuk <grzegorz.pawelczuk@nftlearning.com>
"""
import pytest
import responses
from requests import HTTPError

from nftl_ed_lms_tools.client import EdClient
from nftl_ed_lms_tools.ed_api import EdApi, CallType


@pytest.fixture
def ed_client():
    return EdClient('token')


@pytest.fixture
def broken_url_ed_client():
    return EdClient('token', 'https://example.com/api')


def test_ed_users(broken_url_ed_client):
    # given:
    api1 = broken_url_ed_client.get_users_api()
    api2 = broken_url_ed_client.get_users_api()

    # then:
    assert api1 == api2


def test_ed_user_groups(broken_url_ed_client):
    # given:
    api1 = broken_url_ed_client.get_usergroups_api()
    api2 = broken_url_ed_client.get_usergroups_api()

    # then:
    assert api1 == api2


@responses.activate
def test_ed_get_users(ed_client):
    # given:
    responses.add(
        method=responses.GET,
        url='https://rest.edapp.com/v1/users',
        body='[{"user": "user"}]',
        status=200,
        content_type='application/json'
    )

    # then:
    assert ed_client.get_users_api().get_users()


@responses.activate
def test_ed_get_users_invalid(ed_client):
    # given:
    responses.add(
        method=responses.GET,
        url='https://rest.edapp.com/v1/users',
        body='some <b>html</b>',
        status=200,
        content_type='application/json'
    )

    # then:
    assert not ed_client.get_users_api().get_users()


def test_ed_create_or_update_user_invalid(ed_client, caplog):
    # when :
    with pytest.raises(AttributeError):
        ed_client.get_users_api().create_or_update_user({})

    # then:
    assert 'not provided in user data' in caplog.text


@responses.activate
def test_ed_create_or_update_user(ed_client, caplog):
    # given:
    responses.add(
        method=responses.POST,
        url='https://rest.edapp.com/v1/users',
        body='{"user": "user"}',
        status=200,
        content_type='application/json'
    )

    # then:
    assert ed_client.get_users_api().create_or_update_user({
        'externalIdentifier': 'irrelevant'
    })


@responses.activate
def test_ed_external_token(ed_client):
    # given:
    user = 'irrelevant'
    responses.add(
        method=responses.GET,
        url=f'https://rest.edapp.com/v1/users/external/{user}/token',
        body='{"user": "user"}',
        status=200,
        content_type='application/json'
    )

    # then:
    assert ed_client.get_users_api().get_external_token(user)


@responses.activate
def test_ed_token(ed_client):
    # given:
    user = 'irrelevant'
    responses.add(
        method=responses.GET,
        url=f'https://rest.edapp.com/v1/users/id/{user}/token',
        body='{"user": "user"}',
        status=200,
        content_type='application/json'
    )

    # then:
    assert ed_client.get_users_api().get_token(user)


def test_edapi_response(mocker, caplog):
    # given:
    response = mocker.patch('requests.get').return_value
    response.ok = False
    response.status_code = 500

    # when :
    EdApi._get_data(response)

    # then :
    assert 'HTTP error' in caplog.text


@responses.activate
def test_api_delete(mocker, caplog):
    # given:
    responses.add(responses.GET, 'https://example.com/test',
                  body=HTTPError('Error'))
    api = EdApi('token', 'https://example.com/')

    # when:
    call = api._call('test', None, call_type=CallType.DELETE)

    # then:
    assert not call
    assert 'Error handling req.' in caplog.text


def test_ed_create_or_update_group_invalid(ed_client, caplog):
    # when :
    with pytest.raises(AttributeError):
        ed_client.get_usergroups_api().create_or_update_group({})

    # then:
    assert 'ED group "id" not provided in' in caplog.text


@responses.activate
def test_ed_create_or_update_group(ed_client, caplog):
    # given:
    responses.add(
        method=responses.POST,
        url='https://rest.edapp.com/v1/usergroups',
        body='{"id": "irrelevant"}',
        status=200,
        content_type='application/json'
    )

    # then:
    assert ed_client.get_usergroups_api().create_or_update_group({
        'id': 'irrelevant'
    })


@responses.activate
def test_get_groups(ed_client, caplog):
    # given:
    responses.add(
        method=responses.GET,
        url='https://rest.edapp.com/v1/usergroups',
        body='[{"id": "irrelevant"}]',
        status=200,
        content_type='application/json'
    )

    # then:
    assert ed_client.get_usergroups_api().get()



# -*- coding: utf-8 -*-
"""
Tools for interacting with Ed API.

..codeauthor Grzegorz Pawełczuk <grzegorz.pawelczuk@nftlearning.com>
"""
import logging
from typing import Optional, Any, List, Dict

from nftl_ed_lms_tools.ed_api import EdApi, CallType


class Users(EdApi):
    """
    Ed API wrapper that is handling users category
    """
    API_VERSION = "v1"

    def __init__(self, token: str, api_url: str = None) -> None:
        super().__init__(token, api_url)

    def get_users(self, token=None):
        # type: (Optional[str]) -> Optional[List[Any]]
        """
        Get a list of users.
        http://developer.edapp.com/#users_get
        Args:
            token: optional auth token that will overwrite EdApi token
        Returns:
            None if data is invalid or users info as :py:class:`typing.List`
            when success
        """
        return super()._call(
            method='users',
            token=token,
            call_type=CallType.GET
        )

    def create_or_update_user(self, user_data, token=None):
        # type: (Dict[str,Any], Optional[str]) -> Optional[Dict[str,Any]]
        """
        Create or update a user.
        http://developer.edapp.com/#users_post

        Args:
            user_data: user data to create or update
            token: optional auth token that will overwrite EdApi token
        Returns:
            None if data is invalid or users info as :py:class:`typing.List`
            when success
        """
        if 'externalIdentifier' not in user_data:
            msg = '"username" or "email" not provided in user data!'
            logging.error(msg)
            raise AttributeError(msg)

        return super()._call(
            method='users',
            data=user_data,
            token=token,
            call_type=CallType.POST
        )

    def get_external_token(self, external_id, token=None):
        # type: (str, Optional[str]) -> Optional[Dict[str,Any]]
        """
        Get users authentication token by externalId.
        http://developer.edapp.com/#users_external__id__token_get

        Args:
            external_id: users external identifier
            token: optional auth token that will overwrite EdApi token
        Returns:
            None if data is invalid or users info as :py:class:`typing.List`
            when success
        """
        return super()._call(
            method=f'users/external/{external_id}/token',
            token=token,
            call_type=CallType.GET
        )

    def get_token(self, user_id, token=None):
        # type: (str, Optional[str]) -> Optional[Dict[str,Any]]
        """
        Get users authentication token by id.
        http://developer.edapp.com/#users__id__token_get

        Args:
            user_id: users external identifier
            token: optional auth token that will overwrite EdApi token
        Returns:
            None if data is invalid or users info as :py:class:`typing.List`
            when success
        """
        return super()._call(
            method=f'users/id/{user_id}/token',
            token=token,
            call_type=CallType.GET
        )
